"""
Multitype2 data type class
"""

# pylint: disable=too-few-public-methods

import nadap.schema
from nadap.doc import TableRow
import nadap.types.base
from nadap.base import ValEnv
from nadap.errors import SchemaDefinitionError, DataValidationError

DOC_DT_NAME = "Multitype2"
DOC_DT_DESCRIPTION = """
A **multitype2** data type represents a selection of different data types.
It test data for being a valid input for a listed data type definition -
one by one. If a data type definition matches, a requested preprocessing
is applied. If no data type definition matches, the multitype2 data type
doesn't match and validation fails.

> Note: **multitype2** relaxes restrictions of **multitype** but loses
> details why a validation fails.
"""
DOC_DT_FEATURES = """
- Overlapping python class matching among listed data types is supported.
- Nesting multitype or multitype2 in multitype2 is supported.
- Data type conversion is not supported.
  Use conversion within listed data types instead.
"""
DOC_DT_YAML_EXAMPLE = """
type: multitype2
description: "Example multitype2 definition"
default_value: 0

types:
  - type: int
    reference: ref_key
  - type: str
    minimum: 4
    not_allowed_values:
      - 'test'
    reference: ref_key
  - type: str
    minimum: 2
    not_allowed_values:
      - 'me'
    reference: ref_key
"""


class MultiType2(nadap.types.base.BaseType):
    """
    Multitype2 datatype class
    """

    data_type_name = "multitype2"
    _convert_to_classes = {}

    def __init__(self, **kwargs):
        self.types = []
        super().__init__(**kwargs)

    def _pop_options(self, definition: dict, schema_path: str):
        if "types" not in definition:
            raise SchemaDefinitionError("Option types is required", schema_path)
        self.types = definition.pop("types")
        super()._pop_options(definition, schema_path)

    def _validate_options(self, schema_path: str):
        types_definition = self.types
        self.types = []
        nadap.schema.is_non_empty_list(types_definition, f"{schema_path}.types")
        _data_types = []
        for index, _type in enumerate(types_definition):
            i_path = f"{schema_path}.types[{index}]"
            if isinstance(_type, str):
                _type = {"type": _type}
            nadap.schema.is_dict(_type, i_path)
            dt = self._schema.load_data_type_by_definition(
                definition=_type, path=i_path
            )
            for python_class in dt.python_classes:
                if python_class in self.types:
                    raise SchemaDefinitionError(
                        f"Data type '{dt.data_type_name}' interferes with '"
                        + f"{self.types[python_class].data_type_name}'",
                        i_path,
                    )
            self.types.append(dt)
            _data_types.append(dt.data_type_name)
            for class_ in dt.python_classes:
                if class_ not in self.python_classes:
                    self.python_classes.append(class_)
        super()._validate_options(schema_path=schema_path)

    def _validate_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> "any":
        len_findings = len(env.findings)
        for data_type in self.types:
            try:
                return data_type.validate(data=data, path=path, env=env)
            except DataValidationError:
                if len_findings < len(env.findings):
                    del env.findings[len_findings - len(env.findings) :]
        self._create_finding_with_error(
            msg="None of the data types matches", path=path, env=env
        )
        return data

    @property
    def doc_table_rows(self) -> "list[TableRow]":
        """
        Get doc table rows for this data type definition
        """
        ret_list = super().doc_table_rows
        for index, dt in enumerate(self.types):
            dt_entries = dt.doc_table_rows
            dt_entries[0].variable.name = f"option {index + 1}"
            dt_entries[0].variable.placeholder = True
            for entry in dt_entries:
                entry.skip_yaml = True
            ret_list += dt_entries
            if index + 1 < len(self.types):
                or_entry = TableRow(variable="~or~")
                or_entry.skip_yaml = True
                ret_list.append(or_entry)
        return ret_list

    @property
    def doc_yaml(self) -> "list[str]":
        """
        Get data structure
        """
        if len(self.types) == 1:
            return list(self.types)[0].doc_yaml
        if dict in self.python_classes or list in self.python_classes:
            ret_list = ["# Multitype!!!"]
            for index, dt in enumerate(self.types):
                ret_list += dt.doc_yaml
                if index + 1 < len(self.types):
                    ret_list.append("# or")
            return ret_list
        types = [f"<{x.__name__}>" for x in self.python_classes]
        return [" or ".join(sorted(types))]

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **types** | <code>list[dict&#124;str]</code> | | required<br>min length: 1 | "
            + "List of type definitions |",
            "| &nbsp;&nbsp;- <type> | <code>dict&#124;str</code> | | | |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;*option 1* | <code>str</code> | | "
            + "Data type string (short definition) |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;*option 2* | <code>dict</code> | | "
            + "Data type definition |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "",
            "types:",
            "  - <dict|str>",
            "    # Mulitype!!!",
            "    <str>",
            "    # or",
            "    <dict>",
        ]


DOC_DT_CLASS = MultiType2  # pylint: disable=invalid-name



"""
16bit unsigned Integer data type classes
"""

# pylint: disable=too-few-public-methods

import nadap.types.base
import nadap.mixin.not_allowed_value
import nadap.mixin.min_max
import nadap.mixin.ranges
import nadap.mixin.allowed_value
from nadap.base import ValEnv
from nadap.errors import SchemaDefinitionError

DOC_DT_NAME = "16bit unsigned Integer"
DOC_DT_DESCRIPTION = f"""
An **uint16** data type tests data for being an instance of
python's built-in class `int` and has a value within 0 and {2**16 - 1}.

> If using *uint16* data type with allowed values and no conversion defined,
> check *enum* data type as a better option.
"""
DOC_DT_FEATURES = f"""
- Validate minimal and maximal data value{nadap.mixin.ranges.DOC_FEATURES}
- Validate against allowed and not allowed integer values
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = """
type: uint16
description: "Example uint16 definition"

minimum: 1
maximum: 10000

not_allowed_values:
  - 1002
  - 1003

reference:
  key: ref_key
  mode: consumer
"""


class Uint16(
    nadap.mixin.not_allowed_value.NotAllowedValueMixin,
    nadap.mixin.allowed_value.AllowedValueMixin,
    nadap.mixin.ranges.ValueRangesMixin,
    nadap.mixin.min_max.MinMaxValueMixin,
    nadap.types.base.BaseType,
):
    """
    16-bit unsigned Integer datatype class
    """

    data_type_name = "uint16"
    _cls_python_classes = [int]
    _convert_to_classes = {"str": str, "float": float}
    _doc_data_type = "int"
    _minimum_default = 0
    _maximum_default = 65535

    def _test_data_type(self, data: any, path: str, env: "ValEnv" = None) -> "any":
        data = super()._test_data_type(data=data, path=path, env=env)
        if not 0 <= data < 2**16:
            msg = "Data is out of range of 'uint16'"
            if env is None:
                raise SchemaDefinitionError(msg, path)
            self._create_finding_with_error(msg, path, env)
        return data


DOC_DT_CLASS = Uint16  # pylint: disable=invalid-name



"""
String data type class
"""

# pylint: disable=too-few-public-methods

from typing import TYPE_CHECKING

import nadap.types.base
import nadap.mixin.regex_mode
import nadap.mixin.min_max
import nadap.mixin.replace_empty_to
import nadap.schema
from nadap.base import ValEnv
from nadap.doc import UnorderedTextList

if TYPE_CHECKING:
    from nadap.doc import TextField

DOC_DT_NAME = "String"
DOC_DT_DESCRIPTION = """
A **str** data type tests data for being an instance of
python's built-in class `str`.

> If using *str* data type without regex mode, allowed values and
> no conversion defined, check *enum* data type as a better option.
"""
DOC_DT_FEATURES = """
- Validate string length
- Validate against allowed and not allowed strings
- **Regex Mode** validates data strings based on regex patterns.
- Supports **Referencing Feature**. Adds an option to lookup values in other namespaces.
- Data type conversion is not supported
"""
DOC_DT_YAML_EXAMPLE = r"""
type: str
description: "Example str definition with regex mode disabled"
default_value: "Hello World!"

minimum: 3
maximum: 15

not_allowed_values:
  - nadap sucks!

reference: ref_key
```

Example with regex mode enabled:

```yaml
type: str
description: "Example str definition with regex mode enabled and namespace lookup"

minimum: 4
maximum: 15

regex_mode: true
allowed_values:
  - '[a-z_\.]*'
not_allowed_values:
  - \d+.*

reference:
  key: ref_key
  mode: consumer
  namespace_separator_char: '.'
"""


class Str(
    nadap.mixin.replace_empty_to.ReplaceEmptyToMixin,
    nadap.mixin.regex_mode.RegexModeMixin,
    nadap.mixin.min_max.MinMaxLengthMixin,
    nadap.types.base.BaseType,
):
    """
    String datatype class
    """

    data_type_name = "str"
    _cls_python_classes = [str]
    _convert_to_classes = {}
    _support_ns_separator = True
    _support_replace_empty_to = True
    _doc_data_type = "str"

    def __init__(self, **kwargs):
        self.allowed_values = None
        self.not_allowed_values = []
        super().__init__(**kwargs)

    def _validate_allowed_values_option(self, schema_path: str):
        if self.allowed_values is None:
            return
        regex_objects = []
        av_path = f"{schema_path}.allowed_values"
        nadap.schema.is_non_empty_list(self.allowed_values, av_path)
        for index, value in enumerate(self.allowed_values):
            av_i_path = f"{av_path}[{index}]"
            self._test_data_type(value, av_i_path)
            if self.regex_mode:
                regex_objects.append(
                    nadap.schema.compile_regex_string(
                        pattern=value,
                        multiline=self.regex_multiline,
                        fullmatch=self.regex_fullmatch,
                        schema_path=av_i_path,
                    )
                )
        if self.regex_mode:
            self.allowed_values = regex_objects

    def _validate_not_allowed_values_option(self, schema_path: str):
        regex_objects = []
        av_path = f"{schema_path}.not_allowed_values"
        nadap.schema.is_list(self.not_allowed_values, av_path)
        for index, value in enumerate(self.not_allowed_values):
            av_i_path = f"{av_path}[{index}]"
            self._test_data_type(value, av_i_path)
            if self.regex_mode:
                regex_objects.append(
                    nadap.schema.compile_regex_string(
                        pattern=value,
                        multiline=self.regex_multiline,
                        fullmatch=self.regex_fullmatch,
                        schema_path=av_i_path,
                    )
                )
        if self.regex_mode:
            self.not_allowed_values = regex_objects

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)
        self._validate_allowed_values_option(schema_path)
        self._validate_not_allowed_values_option(schema_path)

    def _pop_options(self, definition: dict, schema_path: str):
        self.allowed_values = definition.pop("allowed_values", None)
        self.not_allowed_values = definition.pop("not_allowed_values", [])
        super()._pop_options(definition, schema_path)

    def _validate_allowed_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ):
        if not self.allowed_values:
            return
        if self.regex_mode:
            matches = nadap.schema.match_regex_objects(self.allowed_values, data)[0]
            if not matches:
                self._create_finding_with_error(
                    msg="Data does not match any allowed regex patterns",
                    path=path,
                    env=env,
                )
        elif data not in self.allowed_values:
            self._create_finding_with_error(
                msg="Data not within allowed strings",
                path=path,
                env=env,
            )

    def _validate_not_allowed_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ):
        if not self.not_allowed_values:
            return
        if self.regex_mode:
            matches, pattern = nadap.schema.match_regex_objects(
                self.not_allowed_values, data
            )
            if matches:
                self._create_finding_with_error(
                    msg=f"Data does match not allowed regex pattern '{pattern.pattern}'",
                    path=path,
                    env=env,
                )
        elif data in self.not_allowed_values:
            self._create_finding_with_error(
                msg="Data is a not allowed string",
                path=path,
                env=env,
            )

    def _validate_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> "any":
        data = super()._validate_data(data=data, path=path, env=env)
        self._validate_allowed_data(data=data, path=path, env=env)
        self._validate_not_allowed_data(data=data, path=path, env=env)
        return data

    @property
    def restrictions(self) -> "TextField":
        """
        Get all restrictions for valid data
        """
        tf = super().restrictions
        if self.regex_mode:
            if self.allowed_values:
                tf.append("allowed regex patterns:")
                tf.append(
                    UnorderedTextList(
                        [rf"`{x.pattern.pattern}`" for x in self.allowed_values]
                    )
                )
            if self.not_allowed_values:
                tf.append("not allowed regex patterns:")
                tf.append(
                    UnorderedTextList(
                        [rf"`{x.pattern.pattern}`" for x in self.not_allowed_values]
                    )
                )
            if self.regex_fullmatch or self.regex_multiline:
                r_options = []
                if self.regex_fullmatch:
                    r_options.append("fullmatch")
                if self.regex_multiline:
                    r_options.append("multiline")
                tf.append(f"Regex match options: {', '.join(r_options)}")
        else:
            if self.allowed_values:
                tf.append("allowed strings:")
                tf.append(UnorderedTextList([rf"`{x}`" for x in self.allowed_values]))
            if self.not_allowed_values:
                tf.append("not allowed strings:")
                tf.append(
                    UnorderedTextList([rf"`{x}`" for x in self.not_allowed_values])
                )
        return tf

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **allowed_values** | <code>list[str]</code> | | | "
            + " Data must match one of these strings/regex strings |",
            "| &nbsp;&nbsp;- < str > | <code>str</code> | | | |",
            "| **not_allowed_values** | <code>list[str]</code> | | | "
            + " Data mustn't match any of these strings/regex strings |",
            "| &nbsp;&nbsp;- < str > | <code>str</code> | | | |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "allowed_values:",
            "  - <str>",
            "not_allowed_values:",
            "  - <str>",
        ]


DOC_DT_CLASS = Str  # pylint: disable=invalid-name



"""
ReplaceEmptyToMixin
"""

# pylint: disable=too-few-public-methods

from nadap.base import UNDEFINED


class ReplaceEmptyToMixin:
    """
    Add option to replace empty data with defined value
    """

    def __init__(self, **kwargs):
        self._replace_empty_to = UNDEFINED
        super().__init__(**kwargs)

    def _pop_options(self, definition: dict, schema_path: str):
        self._replace_empty_to = definition.pop("replace_empty_to", UNDEFINED)
        super()._pop_options(definition=definition, schema_path=schema_path)

    def _convert_data(self, data: any) -> any:
        if self._replace_empty_to is not UNDEFINED and not data:
            return self._replace_empty_to
        return super()._convert_data(data)

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **replace_empty_to** | <code>any</code> | | | "
            + " An empty input value will be replaced by this data |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "replace_empty_to: <any>",
        ]



"""
List classes
"""

# pylint: disable=too-few-public-methods

import nadap.types.base
import nadap.mixin.replace_empty_to
import nadap.mixin.min_max
import nadap.mixin.allow_duplicate
from nadap.doc import TableRow

DOC_DT_NAME = "List"
DOC_DT_DESCRIPTION = """
An **list** data type tests data for being an instance of
python's built-in class `list`.
"""
DOC_DT_FEATURES = """
- Validate list length
- Validate uniqueness of elements
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = """
type: list
description: "Example list definition for a list of strings"
default_value:
  - nadap
  - rulez!

minimum: 2
maximum: 15
allow_duplicates: false

elements: str

reference: ref_key
"""


class ListBase(
    nadap.mixin.replace_empty_to.ReplaceEmptyToMixin,
    nadap.mixin.min_max.MinMaxLengthMixin,
    nadap.types.base.BaseType,
):
    """
    List datatype base class
    """

    _cls_python_classes = [list]
    _support_replace_empty_to = True

    def __init__(self, **kwargs):
        self.elements_data_type = None
        super().__init__(**kwargs)

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)
        self.elements_data_type = self._schema.load_data_type_by_definition(
            self.elements_data_type, f"{schema_path}.elements"
        )

    def _pop_options(self, definition: dict, schema_path: str):
        self.elements_data_type = definition.pop("elements", "any")
        super()._pop_options(definition, schema_path)

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **elements** | <code>dict&#124;str</code> | | required | "
            + "List element's data type definition |",
            "| &nbsp;&nbsp;*option 1* | <code>str</code> | | "
            + "List element's data type string (short definition) |",
            "| &nbsp;&nbsp;*option 2* | <code>dict</code> | | "
            + "List element's data type definition |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "",
            "elements: <dict|str>",
            "  # Mulitype!!!",
            "  <str>",
            "  # or",
            "  <dict>",
        ]


class List(nadap.mixin.allow_duplicate.AllowDuplicateMixin, ListBase):
    """
    List datatype class
    """

    data_type_name = "list"
    _doc_data_type = "list"

    @property
    def yaml_data_type(self) -> str:
        """Get YAML data type string"""
        if self.elements_data_type.python_classes:
            return f"list[{self.elements_data_type.yaml_data_type}]"
        return f"list[{self.elements_data_type.data_type_name}]"

    @property
    def doc_table_rows(self) -> "list[TableRow]":
        """
        Get doc table entries for this data type definition
        """
        e_entries = self.elements_data_type.doc_table_rows
        for e in e_entries:
            e.path = "[]" + e.path
        if len(e_entries) == 1:
            # Element is not a data structure
            e_entries[0].indent = 1
            e_entries[0].variable = self.elements_data_type.doc_value_name
            e_entries[0].variable.placeholder = True
        elif e_entries[0].type.types == ["list"]:
            # Element is a list
            for entry in e_entries:
                entry.indent.level += 1
        else:
            # Element is a non-empty dictionary
            e_entries = e_entries[1:]
            for entry in e_entries:
                entry.indent.level += 1
        e_entries[0].indent.start_list = True

        return super().doc_table_rows + e_entries

    @property
    def doc_yaml(self) -> "list[str]":
        """
        Get data structure
        """
        element_doc_yaml = self.elements_data_type.doc_yaml
        ret_lines = []
        for index, line in enumerate(element_doc_yaml):
            if not index:
                ret_lines.append(f"- {line}")
                continue
            ret_lines.append(f"  {line}")
        return ret_lines


DOC_DT_CLASS = List  # pylint: disable=invalid-name



"""
Multitype data type class
"""

# pylint: disable=too-few-public-methods

import nadap.schema
import nadap.types.base
from nadap.base import ValEnv, str_list_out
from nadap.errors import SchemaDefinitionError
from nadap.doc import TableRow

DOC_DT_NAME = "Multitype"
DOC_DT_DESCRIPTION = """
A **multitype** data type represents a selection of different data types.
It test data for being an instance of a python class within the data types
and selects that data type for further data validation.
"""
DOC_DT_FEATURES = """
- Overlapping python class matching among listed data types is not supported.
  I.e. 'dict' and 'idlist'. Both data types tests data for being an instance of `dict`.
- As 'any' data type does not implement a test for a python class or classes it
  is not supported in multitype.
- Nesting multitype or multitype2 in multitype is supported.
- Data type conversion is not supported.
  Use conversion within listed data types instead.
"""
DOC_DT_YAML_EXAMPLE = """
type: multitype
description: "Example multitype definition"
default_value: 0

types:
  - 'none'
  - type: int
    reference: ref_key
  - type: str
    minimum: 2
    not_allowed_values:
      - test
    reference: ref_key
"""


class MultiType(nadap.types.base.BaseType):
    """
    Multitype datatype class
    """

    data_type_name = "multitype"
    _convert_to_classes = {}

    def __init__(self, **kwargs):
        self.types = {}  # python class -> nadap data type object
        super().__init__(**kwargs)

    def _pop_options(self, definition: dict, schema_path: str):
        if "types" not in definition:
            raise SchemaDefinitionError("Option types is required", schema_path)
        self.types = definition.pop("types")

        super()._pop_options(definition, schema_path)

    def _validate_options(self, schema_path: str):
        types_definition = self.types
        self.types = {}
        nadap.schema.is_non_empty_list(types_definition, f"{schema_path}.types")
        _data_types = []
        for index, _type in enumerate(types_definition):
            i_path = f"{schema_path}.types[{index}]"
            if isinstance(_type, str):
                _type = {"type": _type}
            nadap.schema.is_dict(_type, i_path)
            dt = self._schema.load_data_type_by_definition(
                definition=_type, path=i_path
            )
            if not dt.python_classes:
                raise SchemaDefinitionError(
                    f"Data type '{dt.data_type_name}' not supported within multitype",
                    i_path,
                )
            for python_class in dt.python_classes:
                if python_class in self.types:
                    raise SchemaDefinitionError(
                        f"Data type '{dt.data_type_name}' interferes with '"
                        + f"{self.types[python_class].data_type_name}'",
                        i_path,
                    )
                self.types[python_class] = dt
            _data_types.append(dt.data_type_name)
            self.python_classes.extend(dt.python_classes)
        super()._validate_options(schema_path=schema_path)

    @property
    def _data_type_mismatch_message(self) -> str:
        return "Data is not an instance of " + str_list_out(
            sorted([x.__name__ for x in self.types])
        )

    def _validate_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> "any":
        data_type_found = False
        for match_data_types, data_type in self.types.items():
            if isinstance(data, match_data_types):
                data_type_found = True
                data = data_type.validate(data=data, path=path, env=env)
                break
        if not data_type_found:
            self._create_finding_with_error(
                msg=self._data_type_mismatch_message,
                path=path,
                env=env,
            )
        return data

    @property
    def doc_table_rows(self) -> "list[TableRow]":
        """
        Get doc table rows for this data type definition
        """
        ret_list = super().doc_table_rows
        for index, dt in enumerate(self.types.values()):
            dt_entries = dt.doc_table_rows
            dt_entries[0].variable.name = f"option {index + 1}"
            dt_entries[0].variable.placeholder = True
            for entry in dt_entries:
                entry.skip_yaml = True
            ret_list += dt_entries
            if index + 1 < len(self.types.values()):
                or_entry = TableRow(variable="~or~")
                or_entry.skip_yaml = True
                ret_list.append(or_entry)
        return ret_list

    @property
    def doc_yaml(self) -> "list[str]":
        """
        Get data structure
        """
        if len(self.types) == 1:
            return list(self.types.values())[0].doc_yaml
        if dict in self.python_classes or list in self.python_classes:
            ret_list = ["# Multitype!!!"]
            for index, dt in enumerate(self.types.values()):
                ret_list += dt.doc_yaml
                if index + 1 < len(self.types.values()):
                    ret_list.append("# or")
            return ret_list
        types = [f"<{x.__name__}>" for x in self.python_classes]
        return [" or ".join(sorted(types))]

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **types** | <code>list[dict&#124;str]</code> | | required<br>min length: 1 | "
            + "List of type definitions |",
            "| &nbsp;&nbsp;- <type> | <code>dict&#124;str</code> | | | |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;*option 1* | <code>str</code> | | "
            + "Data type string (short definition) |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;*option 2* | <code>dict</code> | | "
            + "Data type definition |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "",
            "types:",
            "  - <dict|str>",
            "    # Mulitype!!!",
            "    <str>",
            "    # or",
            "    <dict>",
        ]


DOC_DT_CLASS = MultiType  # pylint: disable=invalid-name



"""
IP Address Data Type Class
"""

# pylint: disable=too-few-public-methods

import ipaddress

import nadap.mixin.ip
import nadap.mixin.ranges
import nadap.types.base

DOC_DT_NAME = "IP Address"
DOC_DT_DESCRIPTION = """
An **ip_address** data type tests data for being an instance of
python's built-in class `str` and the string represents an IPv4
or IPv6 address.
"""
DOC_DT_FEATURES = """
- Validate IP specific attributes
- Validate against allowed and not allowed IP address ranges
- Validate against allowed and not allowed IP addresses
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = """
type: ip_address
description: "Example IP address definition"

multicast: False
private: True

not_allowed_values:
  - 10.10.0.1
  - 192.168.0.1
  - 2001:cafe::1

reference:
  key: ref_key
"""


class IpAddress(
    nadap.mixin.ranges.ValueRangesMixin,
    nadap.mixin.ip.IpMixin,
    nadap.types.base.BaseType,
):
    """
    Add handling of strings representing IPv4 or IPv6 addresses
    """

    _ip_data_type_name = "IP address"

    def _get_ip_obj(self, data):
        return ipaddress.ip_address(data)


DOC_DT_CLASS = IpAddress  # pylint: disable=invalid-name



"""
AllowDuplicateMixin
"""

# pylint: disable=too-few-public-methods

from typing import TYPE_CHECKING

from nadap.errors import DataValidationError
from nadap.base import ValEnv
import nadap.results

if TYPE_CHECKING:
    from nadap.doc import TextField


class AllowDuplicateMixin:
    """
    Allow or forbid duplicates in lists
    """

    def __init__(self, **kwargs):
        self.allow_duplicates = True
        super().__init__(**kwargs)

    def _pop_options(self, definition: dict, schema_path: str):
        self.allow_duplicates = definition.pop(
            "allow_duplicates", self.allow_duplicates
        )
        super()._pop_options(definition, schema_path)

    def _validate_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> "any":
        data = super()._validate_data(data=data, path=path, env=env)
        ret_data = []
        invalid = False
        processed_elements = []
        for index, v in enumerate(data):
            if not self.allow_duplicates and v in processed_elements:
                env.findings.append(
                    nadap.results.ValidationFinding(
                        "Duplicate list element", f"{path}[{index}]"
                    )
                )
                invalid = True
            else:
                try:
                    e_data = self.elements_data_type.validate(
                        data=v,
                        path=f"{path}[{index}]",
                        env=env,
                    )
                except DataValidationError:
                    invalid = True
                else:
                    ret_data.append(e_data)
                    processed_elements.append(e_data)
        if invalid:
            raise DataValidationError
        return ret_data

    @property
    def restrictions(self) -> "TextField":
        """
        Get all restrictions for valid data
        """
        tf = super().restrictions
        if not self.allow_duplicates:
            tf.append("all elements must be unique")
        return tf

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **allow_duplicates** | <code>bool</code> | <code>True</code> | | |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "allow_duplicates: <true|false>",
        ]



"""
Central import for types
"""

from nadap.nadap import Nadap
from nadap.errors import SchemaDefinitionError, DataValidationError
from nadap.schema import Schema
from nadap.references import ReferenceFinding
from nadap.results import ValidationFinding
from nadap.base import SET_DEFAULTS, CONVERT_DATA, NOFLAG

__version__ = "2.1.2"
version_info = (2, 1, 2)



"""
4-Byte Value data type class
"""

# pylint: disable=too-few-public-methods

from typing import TYPE_CHECKING

import nadap.types.base
import nadap.mixin.min_max
import nadap.mixin.ranges
import nadap.mixin.allowed_value
import nadap.mixin.not_allowed_value
import nadap.schema
from nadap.base import ValEnv, CONVERT_DATA, str_list_out
from nadap.errors import SchemaDefinitionError
from nadap.doc import UnorderedTextList

if TYPE_CHECKING:
    from nadap.doc import TextField


DOC_DT_NAME = "4-Byte Value"
DOC_DT_DESCRIPTION = """
A **4byte_value** data type tests data for being an instance of
python's built-in class `int` or `str` and if it represents a 4-Byte value.
"""
DOC_DT_FEATURES = """
- Validate integer value
- Validate string in colon-separated format
- Validate string in different dot-separated formats
- Validate hexadecimal string
- Supports format conversion
- Validate minimal and maximal value
- Validate against allowed and not allowed value ranges
- Validate against allowed and not allowed values
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = r"""
type: 4byte_value
description: "Example 4byte_value definition"
default_value: "0:123"

formats:
  - colon-separated
  - 3-dot-separated
  - hex
  - int
convert_to_format: hex

not_allowed_values:
  - 111
  - 1:1
  - 0xff
  - 192.168.1.1

reference: ref_key
"""

CANONICAL = "canonical"
BIT_REVERSED = "bit-reversed"

DOT1_FORMAT = "1-dot-separated"
DOT3_FORMAT = "3-dot-separated"
INT_FORMAT = "int"  # 1003458
COLON_FORMAT = "colon-separated"  # 12:65500
HEX_FORMAT = "hex"  # 0x0012abff or 0x12abff


class Byte4Value(
    nadap.mixin.not_allowed_value.NotAllowedValueMixin,
    nadap.mixin.allowed_value.AllowedValueMixin,
    nadap.mixin.ranges.ValueRangesMixin,
    nadap.mixin.min_max.MinMaxValueMixin,
    nadap.types.base.BaseType,
):
    """
    4-Byte Value datatype class
    """

    data_type_name = "4byte_value"
    _cls_python_classes = [int, str]
    _convert_to_classes = {"int": int, "float": float, "str": str}
    _doc_data_type = "4-Byte Value"

    def __init__(self, **kwargs):
        self._formats = []
        self._detected_format = None
        self._convert_to_format = None
        super().__init__(**kwargs)

    @staticmethod
    def _check_format_str(format_, path):
        if format_ not in [
            DOT1_FORMAT,
            DOT3_FORMAT,
            INT_FORMAT,
            COLON_FORMAT,
            HEX_FORMAT,
        ]:
            raise SchemaDefinitionError(  # pylint: disable=raise-missing-from
                msg="Format unknown",
                path=path,
            )

    def _detect_data_format(self, data) -> str:
        if isinstance(data, int):
            return INT_FORMAT
        if isinstance(data, str):
            if data.count(".") == 1:
                return DOT1_FORMAT
            if data.count(".") == 3:
                return DOT3_FORMAT
            if data.count(":") == 1:
                return COLON_FORMAT
            if data.startswith("0x"):
                return HEX_FORMAT
        return None

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)
        nadap.schema.is_list(self._formats, f"{schema_path}.formats")
        for index, format_ in enumerate(self._formats):
            self._check_format_str(format_, f"{schema_path}.formats[{index}]")
        self._formats = sorted(self._formats)

        if self._convert_to_format is not None:
            nadap.schema.is_str(
                self._convert_to_format, f"{schema_path}.convert_to_format"
            )
            self._check_format_str(
                self._convert_to_format, f"{schema_path}.convert_to_format"
            )

    def _pop_options(self, definition: dict, schema_path: str):
        self._formats = definition.pop("formats", self._formats)
        self._convert_to_format = definition.pop(
            "convert_to_format", self._convert_to_format
        )
        super()._pop_options(definition, schema_path)

    def _validate_format(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> str:
        # pylint: disable=too-many-branches,too-many-statements
        self._detected_format = None
        if not self._formats:
            format_mismatch = "String does not represent a 4-byte value"
        else:
            format_mismatch = (
                "String does not represent a 4-byte value in format "
                + str_list_out(self._formats)
            )
        try:
            r_data = None
            if isinstance(data, int):
                if self._formats and INT_FORMAT not in self._formats:
                    raise ValueError(format_mismatch)
                self._detected_format = INT_FORMAT
                r_data = data
            elif data.count(".") == 1:
                # 1-dot format candidate
                if self._formats and DOT1_FORMAT not in self._formats:
                    raise ValueError(format_mismatch)
                words = data.split(".")
                try:
                    part1 = int(words[0])
                    part2 = int(words[1])
                except ValueError:
                    raise ValueError(  # pylint: disable=raise-missing-from
                        format_mismatch
                    )
                r_data = part1 * 2**16 + part2
                self._detected_format = DOT1_FORMAT
            elif data.count(".") == 3:
                # 3-dot format candidate
                if self._formats and DOT3_FORMAT not in self._formats:
                    raise ValueError(format_mismatch)
                words = data.split(".")
                try:
                    part1 = int(words[0])
                    part2 = int(words[1])
                    part3 = int(words[2])
                    part4 = int(words[3])
                except ValueError:
                    raise ValueError(  # pylint: disable=raise-missing-from
                        format_mismatch
                    )
                r_data = part1 * 2**24 + part2 * 2**16 + part3 * 2**8 + part4
                self._detected_format = DOT3_FORMAT
            elif data.count(":") == 1:
                # colon format candidate
                if self._formats and COLON_FORMAT not in self._formats:
                    raise ValueError(format_mismatch)
                words = data.split(":")
                try:
                    part1 = int(words[0])
                    part2 = int(words[1])
                except ValueError:
                    raise ValueError(  # pylint: disable=raise-missing-from
                        format_mismatch
                    )
                r_data = part1 * 2**16 + part2
                self._detected_format = COLON_FORMAT
            elif data.startswith("0x"):
                # hex format candidate
                if self._formats and HEX_FORMAT not in self._formats:
                    raise ValueError(format_mismatch)
                try:
                    r_data = int(data[2:], 16)
                except ValueError:
                    raise ValueError(  # pylint: disable=raise-missing-from
                        format_mismatch
                    )
                self._detected_format = HEX_FORMAT
            if r_data is not None:
                if not 0 <= r_data < 2**32:
                    raise ValueError("Data out of range of a 4-byte value")
                return r_data
            raise ValueError(format_mismatch)
        except ValueError as e:
            if env is None:
                raise SchemaDefinitionError(  # pylint: disable=raise-missing-from
                    str(e),
                    path,
                )
            self._create_finding_with_error(str(e), path, env)
        return data

    def _apply_data_format(self, data: "any", fmt: str = None) -> "any":
        if fmt == DOT1_FORMAT:
            return f"{int(data / 2**16)}.{data % 2**16}"
        if fmt == DOT3_FORMAT:
            return (
                f"{int(data / 2**24)}.{int((data % 2**24)/ 2**16)}."
                + f"{int((data % 2**16)/ 2**8)}.{data % 2**8}"
            )
        if fmt == COLON_FORMAT:
            return f"{int(data / 2**16)}:{data % 2**16}"
        if fmt == HEX_FORMAT:
            return hex(data)
        return data

    def _preprocess_data(self, data: any, env: "ValEnv"):
        if self._convert_to and self._convert_to in [int, float]:
            return self._convert_to(data)
        if CONVERT_DATA in env.flags:
            if self._convert_to_format is not None:
                data = self._apply_data_format(data, self._convert_to_format)
            else:
                data = self._apply_data_format(data, self._detected_format)
        else:
            data = self._apply_data_format(data, self._detected_format)
        if self._convert_to and self._convert_to == str:
            return self._convert_to(data)
        return data

    def _test_data_type(self, data: str, path: str, env: "ValEnv" = None) -> "str":
        data = super()._test_data_type(data=data, path=path, env=env)
        data = self._validate_format(data=data, path=path, env=env)
        return data

    @property
    def restrictions(self) -> "TextField":
        """
        Get all restrictions for valid data
        """
        tf = super().restrictions
        if self._formats:
            tf.append("4-byte value:")
            tf.append(
                UnorderedTextList(["must be in format " + str_list_out(self._formats)])
            )
        return tf

    @property
    def yaml_data_type(self) -> str:
        """Get YAML data type string"""
        return self._doc_data_type

    @property
    def doc_types(self) -> list[str]:
        """Get list of data type strings"""
        return [self._doc_data_type]

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        allowed_formats = (
            "allowed_values:<br>"
            + f"{cls._markdown_indent}- {INT_FORMAT}<br>"
            + f"{cls._markdown_indent}- {DOT1_FORMAT}<br>"
            + f"{cls._markdown_indent}- {DOT3_FORMAT}<br>"
            + f"{cls._markdown_indent}- {COLON_FORMAT}<br>"
            + f"{cls._markdown_indent}- {HEX_FORMAT}"
        )
        convert_formats = (
            "allowed_values:<br>"
            + f"{cls._markdown_indent}- {DOT1_FORMAT}<br>"
            + f"{cls._markdown_indent}- {DOT3_FORMAT}<br>"
            + f"{cls._markdown_indent}- {COLON_FORMAT}<br>"
            + f"{cls._markdown_indent}- {HEX_FORMAT}"
        )
        f_desc = (
            f"- {INT_FORMAT} example: 101202<br>"
            + f"- {DOT1_FORMAT} example: 123.12<br>"
            + f"- {DOT3_FORMAT} example: 10.1.0.255<br>"
            + f"- {COLON_FORMAT} example: 123:12<br>"
            + f"- {HEX_FORMAT} example: 0x1f23a"
        )
        return super()._doc_options_md_upper_part() + [
            "| **formats** | <code>list[enum]</code> | | | |",
            f"| &nbsp;&nbsp;- < format > | <code>enum</code> | | {allowed_formats} | {f_desc} |",
            f"| **convert_to_format** | <code>enum</code> | | {convert_formats} | |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            f"format: <{INT_FORMAT}|{COLON_FORMAT}|{HEX_FORMAT}>",
            f"convert_to_format: <{COLON_FORMAT}|{HEX_FORMAT}>",
        ]


DOC_DT_CLASS = Byte4Value  # pylint: disable=invalid-name



"""
Mixin for allowed and not allowed value ranges
"""

# pylint: disable=too-few-public-methods

from typing import TYPE_CHECKING

import nadap.schema
from nadap.errors import SchemaDefinitionError
from nadap.base import ValEnv
from nadap.doc import TextList, UnorderedTextList

if TYPE_CHECKING:
    from nadap.doc import TextField

DOC_FEATURES = """
- Validate against allowed and not allowed value ranges"""


class ValueRangesMixin:
    """
    Add allowed and not allowed ranges for value data
    """

    def __init__(self, **kwargs):
        self._allowed_ranges = None
        self._allowed_ranges_fmt = []
        self._not_allowed_ranges = []
        self._not_allowed_ranges_fmt = []
        super().__init__(**kwargs)

    def _validate_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> "any":
        """
        Check if data matches defined data type and apply preprocessing
        """
        data = super()._validate_data(data=data, path=path, env=env)
        if self._allowed_ranges:
            matches = False
            for range_ in self._allowed_ranges:
                if range_[0] <= data <= range_[1]:
                    matches = True
                    break
            if not matches:
                self._create_finding_with_error(
                    msg="Value is not within an allowed range",
                    path=path,
                    env=env,
                )
        for range_ in self._not_allowed_ranges:
            if range_[0] <= data <= range_[1]:
                self._create_finding_with_error(
                    msg="Value is within a not allowed range",
                    path=path,
                    env=env,
                )
        return data

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)

        def _validate_range(
            r_def: dict, path: str
        ) -> "tuple[tuple[any, any], tuple[str, str]]":
            nadap.schema.is_dict(r_def, path)
            start = r_def.pop("start", None)
            end = r_def.pop("end", None)
            nadap.schema.no_more_definition_options(
                definition=r_def, source="range", path=path
            )
            if start is None:
                raise SchemaDefinitionError(msg="Key 'start' missing", path=path)
            if end is None:
                raise SchemaDefinitionError(msg="Key 'end' missing", path=path)
            start_fmt = self._detect_data_format(start)
            start = self._test_data_type(data=start, path=f"{path}.start")
            end_fmt = self._detect_data_format(end)
            end = self._test_data_type(data=end, path=f"{path}.end")
            if start > end:
                raise SchemaDefinitionError(
                    msg="must be greater or equal to 'start'", path=f"{path}.end"
                )
            return ((start, end), (start_fmt, end_fmt))

        if self._allowed_ranges is not None:
            a_path = f"{schema_path}.allowed_ranges"
            nadap.schema.is_non_empty_list(self._allowed_ranges, a_path)
            a_range_def = self._allowed_ranges
            self._allowed_ranges = []
            for index, range_ in enumerate(a_range_def):
                range_values, range_formats = _validate_range(
                    range_, f"{a_path}[{index}]"
                )
                self._allowed_ranges.append(range_values)
                self._allowed_ranges_fmt.append(range_formats)

        if self._not_allowed_ranges:
            n_a_path = f"{schema_path}.not_allowed_ranges"
            nadap.schema.is_list(self._not_allowed_ranges, n_a_path)
            n_a_range_def = self._not_allowed_ranges
            self._not_allowed_ranges = []
            for index, range_ in enumerate(n_a_range_def):
                range_values, range_formats = _validate_range(
                    range_, f"{n_a_path}[{index}]"
                )
                self._not_allowed_ranges.append(range_values)
                self._not_allowed_ranges_fmt.append(range_formats)

    def _pop_options(self, definition: dict, schema_path: str):
        self._allowed_ranges = definition.pop("allowed_ranges", self._allowed_ranges)
        self._not_allowed_ranges = definition.pop(
            "not_allowed_ranges", self._not_allowed_ranges
        )
        super()._pop_options(definition=definition, schema_path=schema_path)

    @property
    def restrictions(self) -> "TextField":
        """
        Get all restrictions for valid data
        """
        tf = super().restrictions
        if self._allowed_ranges is not None:
            tf.append("Allowed value ranges:")
            _restrictions = UnorderedTextList()
            for index, range_ in enumerate(self._allowed_ranges):
                _restrictions.append(
                    TextList(
                        [
                            "start: "
                            + str(
                                self._apply_data_format(
                                    range_[0], self._allowed_ranges_fmt[index][0]
                                )
                            ),
                            "end: "
                            + str(
                                self._apply_data_format(
                                    range_[1], self._allowed_ranges_fmt[index][1]
                                )
                            ),
                        ]
                    )
                )
            tf.append(_restrictions)
        if self._not_allowed_ranges:
            tf.append("Not allowed value ranges:")
            _restrictions = UnorderedTextList()
            for index, range_ in enumerate(self._not_allowed_ranges):
                _restrictions.append(
                    TextList(
                        [
                            "start: "
                            + str(
                                self._apply_data_format(
                                    range_[0], self._not_allowed_ranges_fmt[index][0]
                                )
                            ),
                            "end: "
                            + str(
                                self._apply_data_format(
                                    range_[1], self._not_allowed_ranges_fmt[index][1]
                                )
                            ),
                        ]
                    )
                )
            tf.append(_restrictions)
        return tf

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        data_types = (
            "<code>"
            + "&#124;".join(sorted([x.__name__ for x in cls._cls_python_classes]))
            + "</code>"
        )
        return super()._doc_options_md_upper_part() + [
            "| **allowed_ranges** | <code>list[dict]</code> | | min length: 1 | "
            + "Value must be within defined ranges |",
            f"| {cls._markdown_indent}- **start** | {data_types} | | required |",
            f"| {cls._markdown_indent * 2}**end** | {data_types} | | required |",
            "| **not_allowed_ranges** | <code>list[dict]</code> | | "
            + "| Value mustn't be within defined ranges |",
            f"| {cls._markdown_indent}- **start** | {data_types} | | required |",
            f"| {cls._markdown_indent * 2}**end** | {data_types} | | required |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        data_types = "|".join(sorted([x.__name__ for x in cls._cls_python_classes]))
        return super()._doc_options_yaml_upper_part() + [
            "allowed_ranges:",
            f" - start: <{data_types}>",
            f"   end: <{data_types}>",
            "not_allowed_ranges:",
            f" - start: <{data_types}>",
            f"   end: <{data_types}>",
        ]



"""
Documentation Classes
"""

# pylint: disable=too-few-public-methods

from abc import abstractmethod
from nadap.base import UNDEFINED

HTML_INDENT = "&nbsp;&nbsp;"
START_LIST_CODE = "- "
MARKDOWN_INDENT = "&nbsp;&nbsp;"
STR_INDENT = "  "
NICE_YAML_INDENT = "  "


class TableRowElementMixin:
    """
    Mixin for attributes within the TableRow class
    """

    def __init__(self, *args, **kwargs):
        # pylint: disable=unused-argument
        self.table_row: "TableRow" = None

    @property
    @abstractmethod
    def html(self) -> str:
        """
        Returns a string respresenting this element in html code
        """

    @property
    @abstractmethod
    def markdown(self) -> str:
        """
        Returns a string respresenting this element in markdown code
        """

    @abstractmethod
    def __str__(self):
        return ""


class Indent(TableRowElementMixin):
    """
    Class to handle indent in variable column
    """

    def __init__(self, level: int = 0, start_list: bool = False):
        super().__init__()
        self.level = 0 if level is None else level
        self.start_list = start_list

    def _build_indent_str(self, indent_code: str, start_list_code: str) -> str:
        if self.level < 1:
            return ""
        indent_str = (self.level - 1) * indent_code
        if self.start_list:
            indent_str += start_list_code
        else:
            indent_str += indent_code
        return indent_str

    def __int__(self):
        return self.level

    def __eq__(self, value):
        return self.level == int(value)

    def __ne__(self, value):
        return self.level != int(value)

    def __gt__(self, value):
        return self.level > int(value)

    def __ge__(self, value):
        return self.level >= int(value)

    def __lt__(self, value):
        return self.level < int(value)

    def __le__(self, value):
        return self.level <= int(value)

    def __bool__(self) -> bool:
        return bool(self.level)

    @property
    def html(self) -> str:
        return self._build_indent_str(HTML_INDENT, START_LIST_CODE)

    @property
    def markdown(self) -> str:
        return self.html

    def __str__(self):
        return self._build_indent_str(STR_INDENT, START_LIST_CODE)

    @property
    def yaml(self) -> str:
        """
        Returns a string for indenting in nice YAML text
        """
        return self._build_indent_str(NICE_YAML_INDENT, START_LIST_CODE)


class Variable(TableRowElementMixin):
    """
    Class to handle variable information
    """

    def __init__(self, name: str):
        super().__init__()
        self.name = "" if name is None else name
        self.placeholder = False

    @property
    def _indent_html(self):
        if self.table_row:
            return self.table_row.indent.html
        return ""

    @property
    def _indent_yaml(self):
        if self.table_row:
            return self.table_row.indent.yaml
        return ""

    @property
    def _name_html(self):
        if self.placeholder:
            return f"*<em>{self.name}</em>*"
        return self.name

    @property
    def _name_markdown(self):
        if self.placeholder:
            return rf"\**{self.name}*\*"
        return self.name

    @property
    def _name_yaml(self):
        if self.placeholder:
            return f"*{self.name}*"
        return self.name

    @property
    def html(self) -> str:
        return f"{self._indent_html}{self._name_html}"

    def __str__(self):
        return self.name

    @property
    def markdown(self) -> str:
        return f"{self._indent_html}{self._name_markdown}"

    @property
    def yaml(self) -> str:
        """
        Returns indented name encoded for nice YAML text
        """
        return f"{self._indent_yaml}{self._name_yaml}"


class Type(TableRowElementMixin):
    """
    Class to handle type information
    """

    def __init__(self, t: "str|list[str]"):
        super().__init__()
        if t is None:
            self.types = []
        elif isinstance(t, list):
            self.types = t
        else:
            self.types = [t]

    @property
    def _type_str_html(self) -> str:
        """
        Return a type string encoded as html
        """
        if len(self.types) > 1:
            return "&#124;".join(sorted(self.types))
        return self.types[0]

    @property
    def _type_str_yaml(self) -> str:
        """
        Return a type string encoded as nice YAML text
        """
        if len(self.types) > 1:
            return "|".join(sorted(self.types))
        return self.types[0]

    @property
    def html(self) -> str:
        if not self.types:
            return ""
        return f"<code>{self._type_str_html}</code>"

    def __str__(self):
        return self._type_str_yaml

    @property
    def markdown(self) -> str:
        return self.html

    @property
    def yaml(self) -> str:
        """
        Returns data type encoded for nice YAML text
        """
        if self.types == ["dict"]:
            return "{}"
        if self.types == ["list"]:
            return "[]"
        return f"<{self._type_str_yaml}>"


class DefaultValue(TableRowElementMixin):
    """
    Represent a default value
    """

    def __init__(self, value):
        super().__init__()
        if value is None or value is UNDEFINED:
            self.value = ""
        else:
            self.value = str(value)

    def __bool__(self) -> bool:
        return bool(self.value)

    @property
    def html(self) -> str:
        return self.value

    def __str__(self):
        return self.value

    @property
    def markdown(self) -> str:
        return self.value


def transform_unformatted_text(t: "str|list") -> "TextElement":
    """
    Transform standard python classes to Text classes
    """
    if isinstance(t, str):
        return Text(t)
    return TextList([harmonize_text(x) for x in t])


def harmonize_text(t: "str|list|TextElement") -> "TextElement":
    """
    Transform standard python classes to Text classes if required
    """
    if t is None:
        return Text("")
    if isinstance(t, TextElement):
        return t
    if isinstance(t, (str, list)):
        return transform_unformatted_text(t)
    raise ValueError("Invalid input")


class TextElement:
    """
    Meta class for all TextField elements
    """

    @property
    @abstractmethod
    def html(self) -> str:
        """
        Return text element as html string
        """
        return

    @abstractmethod
    def __str__(self):
        """
        Return text element as plain multiline string
        """
        return ""


class Text(str, TextElement):
    """
    Represents a text in a text field
    """

    @property
    def html(self):
        return self

    def __repr__(self):
        return f"Text({super().__repr__()})"


class TextList(list, TextElement):
    """
    Represents a sequence of text (or other lists) in a text field

    < list element 1 >\n
    < list element 2 >
    """

    def __init__(self, iterable=None):
        if iterable:
            super().__init__([harmonize_text(x) for x in iterable])
        else:
            super().__init__()

    def _html_strings(self, indent: bool = False) -> "list[tuple[str,str]]":
        # pylint: disable=protected-access
        ret_list = []
        html_indent = HTML_INDENT if indent else ""
        for e in self:
            if isinstance(e, TextList):
                ret_list.extend(
                    [(f"{html_indent}{i}", t) for i, t in e._html_strings(indent=True)]
                )
            else:
                ret_list.append((html_indent, e.html))
        return ret_list

    @property
    def html(self):
        return "<br>".join([f"{x[0]}{x[1]}" for x in self._html_strings()])

    def _str_strings(self, indent: bool = False) -> str:
        # pylint: disable=protected-access
        ret_list = []
        str_indent = STR_INDENT if indent else ""
        for e in self:
            if isinstance(e, TextList):
                ret_list.extend(
                    [(f"{str_indent}{i}", t) for i, t in e._str_strings(indent=True)]
                )
            else:
                ret_list.append((str_indent, str(e)))
        return ret_list

    def __str__(self):
        return "\n".join([f"{x[0]}{x[1]}" for x in self._str_strings()])

    def append(self, x):
        x = harmonize_text(x)
        super().append(x)

    def prepend(self, x):
        """
        Insert element at front of the list
        """
        x = harmonize_text(x)
        super().insert(0, x)

    def insert(self, index, x):
        x = harmonize_text(x)
        super().insert(index, x)

    def extend(self, x):
        super().extend([harmonize_text(i) for i in x])

    def __repr__(self):
        return f"TextList({super().__repr__()})"


class UnorderedTextList(TextList):
    """
    Represents an unordered list of text (or other text elements) in a text field

    - < list element 1>\n
    - < list element 2>
    """

    def _html_strings(self, indent: bool = False) -> "list[tuple[str,str]]":
        # pylint: disable=protected-access
        ret_list = []
        for e in self:
            if isinstance(e, TextList):
                e_list = e._html_strings(indent=False)
                if e_list:
                    if e_list[0][0]:
                        if not e_list[0][0].endswith("- "):
                            e_list[0][0] = e_list[0][0][:-2] + "- "
                        ret_list.append((f"{HTML_INDENT}{e_list[0][0]}", e_list[0][1]))
                    else:
                        ret_list.append(("- ", e_list[0][1]))
                    ret_list.extend(
                        [(f"{HTML_INDENT}{x[0]}", x[1]) for x in e_list[1:]]
                    )
            else:
                ret_list.append(("- ", e.html))
        return ret_list

    def _str_strings(self, indent: bool = False) -> str:
        # pylint: disable=protected-access
        ret_list = []
        for e in self:
            if isinstance(e, TextList):
                e_list = e._str_strings(indent=False)
                if e_list:
                    if e_list[0][0]:
                        if not e_list[0][0].endswith(START_LIST_CODE):
                            e_list[0][0] = e_list[0][0][:-2] + START_LIST_CODE
                        ret_list.append((f"{STR_INDENT}{e_list[0][0]}", e_list[0][1]))
                    else:
                        ret_list.append((START_LIST_CODE, e_list[0][1]))
                    ret_list.extend([(f"{STR_INDENT}{x[0]}", x[1]) for x in e_list[1:]])
            else:
                ret_list.append((START_LIST_CODE, str(e)))
        return ret_list

    def __repr__(self):
        return f"UnorderedTextList({super().__repr__()})"


class TextField(TableRowElementMixin):
    """
    Class to handle text information
    """

    def __init__(self, t: "str|list|TextElement|TextField" = None):
        super().__init__()
        self._t = TextList()
        if t is not None:
            self.t = t

    def __bool__(self) -> bool:
        return bool(self._t)

    @property
    def t(self) -> "TextList":
        """
        Get text object
        """
        return self._t

    @t.setter
    def t(self, t: "str|list|TextElement"):
        """
        Set text object
        """
        if t is None:
            self._t = TextList()
        t = self._harmonize_input(t)
        if isinstance(t, TextList):
            self._t = t
        else:
            self._t = TextList()
            self._t.append(t)

    @classmethod
    def _harmonize_input(cls, t: "str|list|TextElement|TextField") -> "TextElement":
        if isinstance(t, TextField):
            return t.t
        return harmonize_text(t)

    def append(self, t: "str|list|TextElement"):
        """
        Add text object to the end of the internal sequence
        """
        if isinstance(t, (TextElement, str)):
            self._t.append(t)
        else:
            self._t.extend(t)

    def prepend(self, t: "str|list|TextElement"):
        """
        Add text object to the front of the internal sequence
        """
        if isinstance(t, list):
            for x in t:
                self._t.prepend(x)
        else:
            self._t.prepend(t)

    @property
    def html(self) -> str:
        """
        Returns text field encoded in html
        """
        return self.t.html

    @property
    def markdown(self) -> str:
        """
        Returns text field encoded in markdown (as html)
        """
        return self.html

    def __str__(self):
        """
        Returns text field as plain multiline string
        """
        return str(self.t)


class TableRow:
    """
    Class represent a row within a documentation table
    """

    # pylint: disable=too-many-instance-attributes,missing-function-docstring

    def __init__(self, **kwargs):
        self.path = kwargs.get("path", "")
        self._indent: "Indent" = None
        self.indent = kwargs.get("indent", None)
        self._variable: "Variable" = None
        self.variable = kwargs.get("variable", None)
        self._type: "Type" = None
        self.type = kwargs.get("type", None)
        self._default: "DefaultValue" = None
        self.default = kwargs.get("default", None)
        self._restrictions: "TextField" = None
        self.restrictions = kwargs.get("restrictions", None)
        self._description: "TextField" = None
        self.description = kwargs.get("description", None)

        # Indicator if corresponding data type is a structure type
        # and sub-sequent table entries exists.
        self.empty_structure = False
        self.key_value = False
        self.skip_yaml = False

    def _setattr(self, attr_, x_, class_):
        if isinstance(x_, class_):
            setattr(self, attr_, x_)
        else:
            setattr(self, attr_, class_(x_))
        getattr(self, attr_).table_row = self

    @property
    def indent(self) -> "Indent":
        return self._indent

    @indent.setter
    def indent(self, x: "Indent|int|None"):
        self._setattr("_indent", x, Indent)

    @property
    def variable(self) -> "Variable":
        return self._variable

    @variable.setter
    def variable(self, x: "Variable"):
        self._setattr("_variable", x, Variable)

    @property
    def type(self) -> "Type":
        return self._type

    @type.setter
    def type(self, x: "Type|str|list[str]|None"):
        self._setattr("_type", x, Type)

    @property
    def default(self) -> "DefaultValue":
        return self._default

    @default.setter
    def default(self, x):
        self._setattr("_default", x, DefaultValue)

    @property
    def description(self) -> "TextField":
        return self._description

    @description.setter
    def description(self, x: "str|list[str]|Text|TextList|UnorderedTextList|TextField"):
        self._setattr("_description", x, TextField)

    @property
    def restrictions(self) -> "TextField":
        return self._restrictions

    @restrictions.setter
    def restrictions(
        self, x: "str|list[str]|Text|TextList|UnorderedTextList|TextField"
    ):
        self._setattr("_restrictions", x, TextField)

    @property
    def markdown(self) -> str:
        """
        Returns a text string representing a row within a markdown table
        """
        return (
            "| "
            + " | ".join(
                [
                    self.variable.markdown,
                    self.type.markdown,
                    self.default.markdown,
                    self.restrictions.markdown,
                    self.description.markdown,
                ]
            )
            + " |"
        )

    @property
    def yaml(self) -> str:
        """
        Returns a text string representing a row within a markdown table
        """
        if (
            not self.empty_structure
            and len(self.type.types) == 1
            and self.type.types[0] in ["dict", "list"]
        ):
            if not self.variable.name:
                return f"{self.indent.yaml}"
            return f"{self.variable.yaml}:"
        if self.variable.placeholder and not self.key_value:
            return self.indent.yaml + self.type.yaml
        return f"{self.variable.yaml}: {self.type.yaml}"


class Doc:
    """
    Class represents the documentation of a schema
    """

    table_columns = [
        "Variable",
        "Type",
        "Default",
        "Restrictions",
        "Description",
    ]

    def __init__(self, rows: "list[TableRow]" = None):
        if rows:
            self.rows: "list[TableRow]" = rows
        else:
            self.rows: "list[TableRow]" = []

    @property
    def markdown_table_header_rows(self) -> list[str]:
        """
        Returns a list of the a markdown table header strings
        """
        return [
            "| " + " | ".join([f"**{x}**" for x in self.table_columns]) + " |",
            "| --- | --- | --- | --- | --- |",
        ]

    def append(self, row: "TableRow"):
        """
        Add a table row at the end of this table
        """
        self.rows.append(row)

    def extend(self, rows: "list[TableRow]"):
        """
        Add a list of table rows at the end of this table
        """
        self.rows.extend(rows)

    @property
    def _markdown_table_rows(self) -> list[str]:
        return [x.markdown for x in self.rows]

    @property
    def markdown_table(self) -> str:
        """
        Returns a multiline string representing a markdown table
        """
        return "\n".join(self.markdown_table_header_rows + self._markdown_table_rows)

    @property
    def nice_yaml(self) -> str:
        """
        Returns a multiline string representing a nice yaml
        """
        return "\n".join([x.yaml for x in self.rows if not x.skip_yaml])
